@@ -79,7 +79,6 @@ uint32_t sem_cpu_sync[CFG_TEE_CORE_NB_CORE];
 DECLARE_KEEP_PAGER(sem_cpu_sync);
 #endif
 
-static void *manifest_dt __nex_bss;
 static unsigned long boot_arg_fdt __nex_bss;
 static unsigned long boot_arg_nsec_entry __nex_bss;
 static unsigned long boot_arg_pageable_part __nex_bss;
@@ -447,7 +446,7 @@ static void print_pager_pool_size(void)
 		stats.npages_all * SMALL_PAGE_SIZE / 1024);
 }
 
-static void init_vcore(tee_mm_pool_t *mm_vcore)
+static void init_virt_pool(tee_mm_pool_t *virt_pool)
 {
 	const vaddr_t begin = VCORE_START_VA;
 	size_t size = TEE_RAM_VA_SIZE;
@@ -458,9 +457,9 @@ static void init_vcore(tee_mm_pool_t *mm_vcore)
 		size = ASAN_MAP_PA - begin;
 #endif
 
-	if (!tee_mm_init(mm_vcore, begin, size, SMALL_PAGE_SHIFT,
+	if (!tee_mm_init(virt_pool, begin, size, SMALL_PAGE_SHIFT,
 			 TEE_MM_POOL_NO_FLAGS))
-		panic("tee_mm_vcore init failed");
+		panic("core_virt_mem_pool init failed");
 }
 
 /*
@@ -615,7 +614,7 @@ static void init_runtime(unsigned long pageable_part)
 	 * Initialize the virtual memory pool used for main_mmu_l2_ttb which
 	 * is supplied to tee_pager_init() below.
 	 */
-	init_vcore(&tee_mm_vcore);
+	init_virt_pool(&core_virt_mem_pool);
 
 	/*
 	 * Assign alias area for pager end of the small page block the rest
@@ -623,9 +622,9 @@ static void init_runtime(unsigned long pageable_part)
 	 * we're guaranteed to not need more than the physical amount of
 	 * TZSRAM.
 	 */
-	mm = tee_mm_alloc2(&tee_mm_vcore,
-			   (vaddr_t)tee_mm_vcore.lo +
-			   tee_mm_vcore.size - TZSRAM_SIZE,
+	mm = tee_mm_alloc2(&core_virt_mem_pool,
+			   (vaddr_t)core_virt_mem_pool.lo +
+			   core_virt_mem_pool.size - TZSRAM_SIZE,
 			   TZSRAM_SIZE);
 	assert(mm);
 	tee_pager_set_alias_area(mm);
@@ -634,7 +633,7 @@ static void init_runtime(unsigned long pageable_part)
 	 * Claim virtual memory which isn't paged.
 	 * Linear memory (flat map core memory) ends there.
 	 */
-	mm = tee_mm_alloc2(&tee_mm_vcore, VCORE_UNPG_RX_PA,
+	mm = tee_mm_alloc2(&core_virt_mem_pool, VCORE_UNPG_RX_PA,
 			   (vaddr_t)(__pageable_start - VCORE_UNPG_RX_PA));
 	assert(mm);
 
@@ -642,7 +641,7 @@ static void init_runtime(unsigned long pageable_part)
 	 * Allocate virtual memory for the pageable area and let the pager
 	 * take charge of all the pages already assigned to that memory.
 	 */
-	mm = tee_mm_alloc2(&tee_mm_vcore, (vaddr_t)__pageable_start,
+	mm = tee_mm_alloc2(&core_virt_mem_pool, (vaddr_t)__pageable_start,
 			   pageable_size);
 	assert(mm);
 	fobj = ro_paged_alloc(mm, hashes, paged_store);
@@ -665,9 +664,10 @@ static void init_runtime(unsigned long pageable_part)
 	 * This setup may happen when a the secure bootloader runs in TZRAM
 	 * and its memory can be reused by OP-TEE once boot stages complete.
 	 */
-	tee_pager_add_pages(tee_mm_vcore.lo,
-			(VCORE_UNPG_RX_PA - tee_mm_vcore.lo) / SMALL_PAGE_SIZE,
-			true);
+	tee_pager_add_pages(core_virt_mem_pool.lo,
+			    (VCORE_UNPG_RX_PA - core_virt_mem_pool.lo) /
+				SMALL_PAGE_SIZE,
+			    true);
 
 	print_pager_pool_size();
 }
@@ -873,133 +873,6 @@ static int config_psci(struct dt_descriptor *dt __unused)
 }
 #endif /*CFG_PSCI_ARM32*/
 
-#ifdef CFG_CORE_DYN_SHM
-static uint64_t get_dt_val_and_advance(const void *data, size_t *offs,
-				       uint32_t cell_size)
-{
-	uint64_t rv = 0;
-
-	if (cell_size == 1) {
-		uint32_t v;
-
-		memcpy(&v, (const uint8_t *)data + *offs, sizeof(v));
-		*offs += sizeof(v);
-		rv = fdt32_to_cpu(v);
-	} else {
-		uint64_t v;
-
-		memcpy(&v, (const uint8_t *)data + *offs, sizeof(v));
-		*offs += sizeof(v);
-		rv = fdt64_to_cpu(v);
-	}
-
-	return rv;
-}
-
-/*
- * Find all non-secure memory from DT. Memory marked inaccessible by Secure
- * World is ignored since it could not be mapped to be used as dynamic shared
- * memory.
- */
-static int get_nsec_memory_helper(void *fdt, struct core_mmu_phys_mem *mem)
-{
-	const uint8_t *prop = NULL;
-	uint64_t a = 0;
-	uint64_t l = 0;
-	size_t prop_offs = 0;
-	size_t prop_len = 0;
-	int elems_total = 0;
-	int addr_size = 0;
-	int len_size = 0;
-	int offs = 0;
-	size_t n = 0;
-	int len = 0;
-
-	addr_size = fdt_address_cells(fdt, 0);
-	if (addr_size < 0)
-		return 0;
-
-	len_size = fdt_size_cells(fdt, 0);
-	if (len_size < 0)
-		return 0;
-
-	while (true) {
-		offs = fdt_node_offset_by_prop_value(fdt, offs, "device_type",
-						     "memory",
-						     sizeof("memory"));
-		if (offs < 0)
-			break;
-
-		if (fdt_get_status(fdt, offs) != (DT_STATUS_OK_NSEC |
-						   DT_STATUS_OK_SEC))
-			continue;
-
-		prop = fdt_getprop(fdt, offs, "reg", &len);
-		if (!prop)
-			continue;
-
-		prop_len = len;
-		for (n = 0, prop_offs = 0; prop_offs < prop_len; n++) {
-			a = get_dt_val_and_advance(prop, &prop_offs, addr_size);
-			if (prop_offs >= prop_len) {
-				n--;
-				break;
-			}
-
-			l = get_dt_val_and_advance(prop, &prop_offs, len_size);
-			if (mem) {
-				mem->type = MEM_AREA_DDR_OVERALL;
-				mem->addr = a;
-				mem->size = l;
-				mem++;
-			}
-		}
-
-		elems_total += n;
-	}
-
-	return elems_total;
-}
-
-static struct core_mmu_phys_mem *get_nsec_memory(void *fdt, size_t *nelems)
-{
-	struct core_mmu_phys_mem *mem = NULL;
-	int elems_total = 0;
-
-	elems_total = get_nsec_memory_helper(fdt, NULL);
-	if (elems_total <= 0)
-		return NULL;
-
-	mem = nex_calloc(elems_total, sizeof(*mem));
-	if (!mem)
-		panic();
-
-	elems_total = get_nsec_memory_helper(fdt, mem);
-	assert(elems_total > 0);
-
-	*nelems = elems_total;
-
-	return mem;
-}
-#endif /*CFG_CORE_DYN_SHM*/
-
-#ifdef CFG_CORE_RESERVED_SHM
-static int mark_static_shm_as_reserved(struct dt_descriptor *dt)
-{
-	vaddr_t shm_start;
-	vaddr_t shm_end;
-
-	core_mmu_get_mem_by_type(MEM_AREA_NSEC_SHM, &shm_start, &shm_end);
-	if (shm_start != shm_end)
-		return add_res_mem_dt_node(dt, "optee_shm",
-					   virt_to_phys((void *)shm_start),
-					   shm_end - shm_start);
-
-	DMSG("No SHM configured");
-	return -1;
-}
-#endif /*CFG_CORE_RESERVED_SHM*/
-
 static int mark_tzdram_as_reserved(struct dt_descriptor *dt)
 {
 	return add_res_mem_dt_node(dt, "optee_core", CFG_TZDRAM_START,
@@ -1031,135 +904,8 @@ static void update_external_dt(void)
 static void update_external_dt(void)
 {
 }
-
-#ifdef CFG_CORE_DYN_SHM
-static struct core_mmu_phys_mem *get_nsec_memory(void *fdt __unused,
-						 size_t *nelems __unused)
-{
-	return NULL;
-}
-#endif /*CFG_CORE_DYN_SHM*/
 #endif /*!CFG_DT*/
 
-#if defined(CFG_CORE_FFA)
-void *get_manifest_dt(void)
-{
-	return manifest_dt;
-}
-
-static void reinit_manifest_dt(void)
-{
-	paddr_t pa = (unsigned long)manifest_dt;
-	void *fdt = NULL;
-	int ret = 0;
-
-	if (!pa) {
-		EMSG("No manifest DT found");
-		return;
-	}
-
-	fdt = core_mmu_add_mapping(MEM_AREA_MANIFEST_DT, pa, CFG_DTB_MAX_SIZE);
-	if (!fdt)
-		panic("Failed to map manifest DT");
-
-	manifest_dt = fdt;
-
-	ret = fdt_check_full(fdt, CFG_DTB_MAX_SIZE);
-	if (ret < 0) {
-		EMSG("Invalid manifest Device Tree at %#lx: error %d", pa, ret);
-		panic();
-	}
-
-	IMSG("manifest DT found");
-}
-
-static TEE_Result release_manifest_dt(void)
-{
-	if (!manifest_dt)
-		return TEE_SUCCESS;
-
-	if (core_mmu_remove_mapping(MEM_AREA_MANIFEST_DT, manifest_dt,
-				    CFG_DTB_MAX_SIZE))
-		panic("Failed to remove temporary manifest DT mapping");
-	manifest_dt = NULL;
-
-	return TEE_SUCCESS;
-}
-
-boot_final(release_manifest_dt);
-#else
-void *get_manifest_dt(void)
-{
-	return NULL;
-}
-
-static void reinit_manifest_dt(void)
-{
-}
-#endif /*CFG_CORE_FFA*/
-
-#ifdef CFG_CORE_DYN_SHM
-static void discover_nsec_memory(void)
-{
-	struct core_mmu_phys_mem *mem;
-	const struct core_mmu_phys_mem *mem_begin = NULL;
-	const struct core_mmu_phys_mem *mem_end = NULL;
-	size_t nelems;
-	void *fdt = get_external_dt();
-
-	if (fdt) {
-		mem = get_nsec_memory(fdt, &nelems);
-		if (mem) {
-			core_mmu_set_discovered_nsec_ddr(mem, nelems);
-			return;
-		}
-
-		DMSG("No non-secure memory found in external DT");
-	}
-
-	fdt = get_embedded_dt();
-	if (fdt) {
-		mem = get_nsec_memory(fdt, &nelems);
-		if (mem) {
-			core_mmu_set_discovered_nsec_ddr(mem, nelems);
-			return;
-		}
-
-		DMSG("No non-secure memory found in embedded DT");
-	}
-
-	mem_begin = phys_ddr_overall_begin;
-	mem_end = phys_ddr_overall_end;
-	nelems = mem_end - mem_begin;
-	if (nelems) {
-		/*
-		 * Platform cannot use both register_ddr() and the now
-		 * deprecated register_dynamic_shm().
-		 */
-		assert(phys_ddr_overall_compat_begin ==
-		       phys_ddr_overall_compat_end);
-	} else {
-		mem_begin = phys_ddr_overall_compat_begin;
-		mem_end = phys_ddr_overall_compat_end;
-		nelems = mem_end - mem_begin;
-		if (!nelems)
-			return;
-		DMSG("Warning register_dynamic_shm() is deprecated, please use register_ddr() instead");
-	}
-
-	mem = nex_calloc(nelems, sizeof(*mem));
-	if (!mem)
-		panic();
-
-	memcpy(mem, phys_ddr_overall_begin, sizeof(*mem) * nelems);
-	core_mmu_set_discovered_nsec_ddr(mem, nelems);
-}
-#else /*CFG_CORE_DYN_SHM*/
-static void discover_nsec_memory(void)
-{
-}
-#endif /*!CFG_CORE_DYN_SHM*/
-
 #ifdef CFG_NS_VIRTUALIZATION
 static TEE_Result virt_init_heap(void)
 {
@@ -1508,7 +1254,7 @@ unsigned long __weak get_aslr_seed(void)
 		goto err;
 	}
 
-	return fdt64_to_cpu(*seed);
+	return fdt64_to_cpu(fdt64_ld(seed));
 
 err:
 	/* Try platform implementation */
@@ -1665,15 +1411,16 @@ void __weak boot_save_args(unsigned long a0, unsigned long a1,
 	if (IS_ENABLED(CFG_CORE_FFA)) {
 		if (IS_ENABLED(CFG_CORE_SEL2_SPMC) ||
 		    IS_ENABLED(CFG_CORE_EL3_SPMC))
-			manifest_dt = get_fdt_from_boot_info((void *)a0);
+			init_manifest_dt(get_fdt_from_boot_info((void *)a0));
 		else
-			manifest_dt = (void *)a0;
+			init_manifest_dt((void *)a0);
 		if (IS_ENABLED(CFG_CORE_SEL2_SPMC) &&
 		    IS_ENABLED(CFG_CORE_PHYS_RELOCATABLE)) {
 			paddr_t base = 0;
 			size_t size = 0;
 
-			get_sec_mem_from_manifest(manifest_dt, &base, &size);
+			get_sec_mem_from_manifest(get_manifest_dt(),
+						  &base, &size);
 			core_mmu_set_secure_memory(base, size);
 		}
 	} else {
