@@ -1,7 +1,7 @@
 // SPDX-License-Identifier: BSD-2-Clause
 /*
  * Copyright (c) 2018, EPAM Systems. All rights reserved.
- * Copyright (c) 2023, Linaro Limited
+ * Copyright (c) 2023-2024, Linaro Limited
  */
 
 #include <bitstring.h>
@@ -10,6 +10,7 @@
 #include <kernel/linker.h>
 #include <kernel/misc.h>
 #include <kernel/mutex.h>
+#include <kernel/notif.h>
 #include <kernel/panic.h>
 #include <kernel/refcount.h>
 #include <kernel/spinlock.h>
@@ -38,6 +39,15 @@ tee_mm_pool_t virt_mapper_pool __nex_bss;
 /* Memory used by OP-TEE core */
 struct tee_mmap_region *kmemory_map __nex_bss;
 
+struct guest_spec_data {
+	size_t size;
+	void (*destroy)(void *data);
+};
+
+static bool add_disabled __nex_bss;
+static unsigned gsd_count __nex_bss;
+static struct guest_spec_data *gsd_array __nex_bss;
+
 struct guest_partition {
 	LIST_ENTRY(guest_partition) link;
 	struct mmu_partition *mmu_prtn;
@@ -48,6 +58,8 @@ struct guest_partition {
 	tee_mm_entry_t *ta_ram;
 	tee_mm_entry_t *tables;
 	bool runtime_initialized;
+	bool got_guest_destroyed;
+	bool shutting_down;
 	uint16_t id;
 	struct refcount refc;
 #ifdef CFG_CORE_SEL1_SPMC
@@ -55,6 +67,7 @@ struct guest_partition {
 	uint8_t cookie_count;
 	bitstr_t bit_decl(shm_bits, SPMC_CORE_SEL1_MAX_SHM_COUNT);
 #endif
+	void **data_array;
 };
 
 struct guest_partition *current_partition[CFG_TEE_CORE_NB_CORE] __nex_bss;
@@ -290,6 +303,40 @@ err:
 	return res;
 }
 
+static void destroy_gsd(struct guest_partition *prtn, bool free_only)
+{
+	size_t n = 0;
+
+	for (n = 0; n < gsd_count; n++) {
+		if (!free_only && prtn->data_array[n] && gsd_array[n].destroy)
+			gsd_array[n].destroy(prtn->data_array[n]);
+		nex_free(prtn->data_array[n]);
+	}
+	nex_free(prtn->data_array);
+	prtn->data_array = NULL;
+}
+
+static TEE_Result alloc_gsd(struct guest_partition *prtn)
+{
+	unsigned int n = 0;
+
+	if (!gsd_count)
+		return TEE_SUCCESS;
+
+	prtn->data_array = nex_calloc(gsd_count, sizeof(void *));
+	if (!prtn->data_array)
+		return TEE_ERROR_OUT_OF_MEMORY;
+
+	for (n = 0; n < gsd_count; n++) {
+		prtn->data_array[n] = nex_calloc(1, gsd_array[n].size);
+		if (!prtn->data_array[n]) {
+			destroy_gsd(prtn, true /*free_only*/);
+			return TEE_ERROR_OUT_OF_MEMORY;
+		}
+	}
+
+	return TEE_SUCCESS;
+}
 TEE_Result virt_guest_created(uint16_t guest_id)
 {
 	struct guest_partition *prtn = NULL;
@@ -300,14 +347,16 @@ TEE_Result virt_guest_created(uint16_t guest_id)
 	if (!prtn)
 		return TEE_ERROR_OUT_OF_MEMORY;
 
+	res = alloc_gsd(prtn);
+	if (res)
+		goto err_free_prtn;
+
 	prtn->id = guest_id;
 	mutex_init(&prtn->mutex);
 	refcount_set(&prtn->refc, 1);
 	res = configure_guest_prtn_mem(prtn);
-	if (res) {
-		nex_free(prtn);
-		return res;
-	}
+	if (res)
+		goto err_free_gsd;
 
 	set_current_prtn(prtn);
 
@@ -326,6 +375,12 @@ TEE_Result virt_guest_created(uint16_t guest_id)
 	core_mmu_set_default_prtn();
 
 	return TEE_SUCCESS;
+
+err_free_gsd:
+	destroy_gsd(prtn, true /*free_only*/);
+err_free_prtn:
+	nex_free(prtn);
+	return res;
 }
 
 static bool
@@ -343,40 +398,97 @@ prtn_have_remaining_resources(struct guest_partition *prtn __maybe_unused)
 #endif
 }
 
-TEE_Result virt_guest_destroyed(uint16_t guest_id)
+static void get_prtn(struct guest_partition *prtn)
+{
+	if (!refcount_inc(&prtn->refc))
+		panic();
+}
+
+uint16_t virt_get_guest_id(struct guest_partition *prtn)
+{
+	if (!prtn)
+		return 0;
+	return prtn->id;
+}
+
+static struct guest_partition *find_guest_by_id_unlocked(uint16_t guest_id)
+{
+	struct guest_partition *prtn = NULL;
+
+	LIST_FOREACH(prtn, &prtn_list, link)
+		if (!prtn->shutting_down && prtn->id == guest_id)
+			return prtn;
+
+	return NULL;
+}
+
+struct guest_partition *virt_next_guest(struct guest_partition *prtn)
+{
+	struct guest_partition *ret = NULL;
+	uint32_t exceptions = 0;
+
+	exceptions = cpu_spin_lock_xsave(&prtn_list_lock);
+	if (prtn)
+		ret = LIST_NEXT(prtn, link);
+	else
+		ret = LIST_FIRST(&prtn_list);
+
+	while (ret && ret->shutting_down)
+		ret = LIST_NEXT(prtn, link);
+	if (ret)
+		get_prtn(ret);
+	cpu_spin_unlock_xrestore(&prtn_list_lock, exceptions);
+
+	virt_put_guest(prtn);
+
+	return ret;
+}
+
+struct guest_partition *virt_get_current_guest(void)
+{
+	struct guest_partition *prtn = get_current_prtn();
+
+	if (prtn)
+		get_prtn(prtn);
+	return prtn;
+}
+
+struct guest_partition *virt_get_guest(uint16_t guest_id)
 {
 	struct guest_partition *prtn = NULL;
 	uint32_t exceptions = 0;
-	bool do_free = true;
-
-	IMSG("Removing guest %d", guest_id);
 
 	exceptions = cpu_spin_lock_xsave(&prtn_list_lock);
-
-	LIST_FOREACH(prtn, &prtn_list, link) {
-		if (prtn->id == guest_id) {
-			if (!refcount_dec(&prtn->refc)) {
-				EMSG("Guest thread(s) is still running. refc = %d",
-				     refcount_val(&prtn->refc));
-				panic();
-			}
-			LIST_REMOVE(prtn, link);
-			if (prtn_have_remaining_resources(prtn)) {
-				LIST_INSERT_HEAD(&prtn_destroy_list, prtn,
-						 link);
-				/*
-				 * Delay the nex_free() until
-				 * virt_reclaim_cookie_from_destroyed_guest()
-				 * is done with this partition.
-				 */
-				do_free = false;
-			}
-			break;
-		}
-	}
+	prtn = find_guest_by_id_unlocked(guest_id);
+	if (prtn)
+		get_prtn(prtn);
 	cpu_spin_unlock_xrestore(&prtn_list_lock, exceptions);
 
-	if (prtn) {
+	return prtn;
+}
+
+void virt_put_guest(struct guest_partition *prtn)
+{
+	if (prtn && refcount_dec(&prtn->refc)) {
+		uint32_t exceptions = 0;
+		bool do_free = true;
+
+		assert(prtn->shutting_down);
+
+		exceptions = cpu_spin_lock_xsave(&prtn_list_lock);
+		LIST_REMOVE(prtn, link);
+		if (prtn_have_remaining_resources(prtn)) {
+			LIST_INSERT_HEAD(&prtn_destroy_list, prtn, link);
+			/*
+			 * Delay the nex_free() until
+			 * virt_reclaim_cookie_from_destroyed_guest()
+			 * is done with this partition.
+			 */
+			do_free = false;
+		}
+		cpu_spin_unlock_xrestore(&prtn_list_lock, exceptions);
+
+		destroy_gsd(prtn, false /*!free_only*/);
 		tee_mm_free(prtn->tee_ram);
 		prtn->tee_ram = NULL;
 		tee_mm_free(prtn->ta_ram);
@@ -389,18 +501,44 @@ TEE_Result virt_guest_destroyed(uint16_t guest_id)
 		prtn->memory_map = NULL;
 		if (do_free)
 			nex_free(prtn);
-	} else
+	}
+}
+
+TEE_Result virt_guest_destroyed(uint16_t guest_id)
+{
+	struct guest_partition *prtn = NULL;
+	uint32_t exceptions = 0;
+
+	IMSG("Removing guest %"PRId16, guest_id);
+
+	exceptions = cpu_spin_lock_xsave(&prtn_list_lock);
+
+	prtn = find_guest_by_id_unlocked(guest_id);
+	if (prtn && !prtn->got_guest_destroyed)
+		prtn->got_guest_destroyed = true;
+	else
+		prtn = NULL;
+
+	cpu_spin_unlock_xrestore(&prtn_list_lock, exceptions);
+
+	if (prtn) {
+		notif_deliver_atomic_event(NOTIF_EVENT_SHUTDOWN, prtn->id);
+
+		exceptions = cpu_spin_lock_xsave(&prtn_list_lock);
+		prtn->shutting_down = true;
+		cpu_spin_unlock_xrestore(&prtn_list_lock, exceptions);
+
+		virt_put_guest(prtn);
+	} else {
 		EMSG("Client with id %d is not found", guest_id);
+	}
 
 	return TEE_SUCCESS;
 }
 
 TEE_Result virt_set_guest(uint16_t guest_id)
 {
-	struct guest_partition *prtn;
-	uint32_t exceptions;
-
-	prtn = get_current_prtn();
+	struct guest_partition *prtn = get_current_prtn();
 
 	/* This can be true only if we return from IRQ RPC */
 	if (prtn && prtn->id == guest_id)
@@ -409,20 +547,14 @@ TEE_Result virt_set_guest(uint16_t guest_id)
 	if (prtn)
 		panic("Virtual guest partition is already set");
 
-	exceptions = cpu_spin_lock_xsave(&prtn_list_lock);
-	LIST_FOREACH(prtn, &prtn_list, link) {
-		if (prtn->id == guest_id) {
-			set_current_prtn(prtn);
-			core_mmu_set_prtn(prtn->mmu_prtn);
-			refcount_inc(&prtn->refc);
-			cpu_spin_unlock_xrestore(&prtn_list_lock,
-						 exceptions);
-			return TEE_SUCCESS;
-		}
-	}
-	cpu_spin_unlock_xrestore(&prtn_list_lock, exceptions);
+	prtn = virt_get_guest(guest_id);
+	if (!prtn)
+		return TEE_ERROR_ITEM_NOT_FOUND;
 
-	return TEE_ERROR_ITEM_NOT_FOUND;
+	set_current_prtn(prtn);
+	core_mmu_set_prtn(prtn->mmu_prtn);
+
+	return TEE_SUCCESS;
 }
 
 void virt_unset_guest(void)
@@ -434,8 +566,7 @@ void virt_unset_guest(void)
 
 	set_current_prtn(NULL);
 	core_mmu_set_default_prtn();
-	if (refcount_dec(&prtn->refc))
-		panic();
+	virt_put_guest(prtn);
 }
 
 void virt_on_stdcall(void)
@@ -619,3 +750,49 @@ TEE_Result virt_reclaim_cookie_from_destroyed_guest(uint16_t guest_id,
 	return res;
 }
 #endif /*CFG_CORE_SEL1_SPMC*/
+
+TEE_Result virt_add_guest_spec_data(unsigned int *data_id, size_t data_size,
+				    void (*data_destroy)(void *data))
+{
+	void *p = NULL;
+
+	/*
+	 * This function only executes successfully in a single threaded
+	 * environment before exiting to the normal world the first time.
+	 * If add_disabled is true, it means we're not in this environment
+	 * any longer.
+	 */
+
+	if (add_disabled)
+		return TEE_ERROR_BAD_PARAMETERS;
+
+	p = nex_realloc(gsd_array, sizeof(*gsd_array) * (gsd_count + 1));
+	if (!p)
+		return TEE_ERROR_OUT_OF_MEMORY;
+	gsd_array = p;
+
+	gsd_array[gsd_count] = (struct guest_spec_data){
+		.size = data_size,
+		.destroy = data_destroy,
+	};
+	*data_id = gsd_count + 1;
+	gsd_count++;
+	return TEE_SUCCESS;
+}
+
+void *virt_get_guest_spec_data(struct guest_partition *prtn,
+			       unsigned int data_id)
+{
+	assert(data_id);
+	if (!data_id || !prtn || data_id > gsd_count)
+		return NULL;
+	return prtn->data_array[data_id - 1];
+}
+
+static TEE_Result virt_disable_add(void)
+{
+	add_disabled = true;
+
+	return TEE_SUCCESS;
+}
+nex_release_init_resource(virt_disable_add);
