@@ -10,6 +10,7 @@
 #include <assert.h>
 #include <compiler.h>
 #include <kernel/user_ta.h>
+#include <mm/tee_mm.h>
 #include <mm/tee_mmu_types.h>
 #include <types_ext.h>
 #include <util.h>
@@ -276,6 +277,15 @@ struct core_mmu_phys_mem {
 #define phys_mem_map_end \
 	SCATTERED_ARRAY_END(phys_mem_map, struct core_mmu_phys_mem)
 
+/* Physical Secure DDR pool */
+extern tee_mm_pool_t tee_mm_sec_ddr;
+
+/* Virtual memory pool for core mappings */
+extern tee_mm_pool_t core_virt_mem_pool;
+
+/* Virtual memory pool for shared memory mappings */
+extern tee_mm_pool_t core_virt_shm_pool;
+
 #ifdef CFG_CORE_RESERVED_SHM
 /* Default NSec shared memory allocated from NSec world */
 extern unsigned long default_nsec_shm_paddr;
