@@ -18,6 +18,7 @@
 #include <kernel/user_mode_ctx.h>
 #include <kernel/virtualization.h>
 #include <libfdt.h>
+#include <memtag.h>
 #include <mm/core_memprot.h>
 #include <mm/core_mmu.h>
 #include <mm/mobj.h>
@@ -35,6 +36,15 @@
 
 #define SHM_VASPACE_SIZE	(1024 * 1024 * 32)
 
+/* Physical Secure DDR pool */
+tee_mm_pool_t tee_mm_sec_ddr;
+
+/* Virtual memory pool for core mappings */
+tee_mm_pool_t core_virt_mem_pool;
+
+/* Virtual memory pool for shared memory mappings */
+tee_mm_pool_t core_virt_shm_pool;
+
 #ifdef CFG_CORE_PHYS_RELOCATABLE
 unsigned long core_mmu_tee_load_pa __nex_bss;
 #else
@@ -277,7 +287,9 @@ static struct tee_mmap_region *find_map_by_pa(unsigned long pa)
 	struct tee_mmap_region *map = get_memory_map();
 
 	while (!core_mmap_is_end_of_table(map)) {
-		if (pa >= map->pa && pa <= (map->pa + map->size - 1))
+		/* Skip unmapped regions */
+		if ((map->attr & TEE_MATTR_VALID_BLOCK) &&
+		    pa >= map->pa && pa <= (map->pa + map->size - 1))
 			return map;
 		map++;
 	}
@@ -997,7 +1009,7 @@ static void collect_device_mem_ranges(struct tee_mmap_region *memory_map,
 		}
 
 		add_phys_mem(memory_map, num_elems, name, MEM_AREA_IO_SEC,
-			     base, base + page_count * SMALL_PAGE_SIZE, last);
+			     base, page_count * SMALL_PAGE_SIZE, last);
 	}
 }
 
@@ -2328,7 +2340,7 @@ paddr_t virt_to_phys(void *va)
 
 	if (!arch_va2pa_helper(va, &pa))
 		pa = 0;
-	check_pa_matches_va(va, pa);
+	check_pa_matches_va(memtag_strip_tag(va), pa);
 	return pa;
 }
 
